from functools import wraps

import falcon
from spectree import SpecTree


spectree = SpecTree('falcon')


@wraps(spectree.validate)
def specification(*args, skip_validation: bool = True, **kwargs):
    return spectree.validate(
        *args,
        skip_validation=skip_validation,
        **kwargs,
    )


@wraps(spectree.register)
def register_specifications(app: falcon.App):
    """
    Регистрация спецификаций SpecTree.
    Регистрацию обязательно проводить ПОСЛЕ регистрации ресурсов в приложении,
    иначе SpecTree не увидит URL-ы ресурсов.
    """
    spectree.register(app)
    return app



from typing import Any

from msgspec import ValidationError
from classic.error_handling import Error, ErrorsList

from falcon import Request, Response
from falcon import status_codes


def validation_error(
    request: Request, response: Response,
    error: ValidationError, params: dict[str, Any],
):
    response.status = status_codes.HTTP_400
    response.media = str(error)


def app_error(
    request: Request, response: Response,
    error: Error, params: dict[str, Any],
):
    response.status = status_codes.HTTP_422
    response.media = [{'type': error.code,
                       'msg': error.message,
                       'ctx': error.context}]


def app_errors_list(
    request: Request, response: Response,
    error: ErrorsList, params: dict[str, Any],
):
    response.status = status_codes.HTTP_422
    response.media = [
        {'type': e.code,
         'msg': e.message,
         'ctx': e.context}
        for e in error.errors
    ]



from functools import wraps
from typing import Any, Callable

from classic.components import add_extra_annotation, doublewrap
from classic.error_handling import Error

from falcon import Request, Response


Factory = Callable[[Any], Any]


class Authenticator:

    def __init__(self, factory: Factory = None):
        self.factory = factory

    def __call__(self, request: Request):
        raise NotImplementedError


class AuthenticationFailed(Error):

    def __init__(self, reason: str = 'Authentication failed'):
        self.message = reason


@doublewrap
def authenticate(method, prop: str = 'authenticator'):
    """

    """

    @wraps(method)
    def wrapper(self, request: Request, *args, **kwargs):
        authenticator = getattr(self, prop)
        try:
            authenticator(request)
        except KeyError:
            raise AuthenticationFailed

        return method(self, request, *args, **kwargs)

    return add_extra_annotation(wrapper, prop, Authenticator)


class AuthenticationMiddleware:
    authenticate: Authenticator

    def __init__(self, authenticator: Authenticator):
        self.authenticate = authenticator

    def process_request(self, request: Request, response: Response):
        self.authenticate(request)



import falcon

from .error_handlers import register_error_handlers
from .specification import register_specifications
from .serialization import register_serializer


def register_all(app: falcon.App):
    """
    Регистрирует весь функционал библиотеки разом:
    - обработчики ошибок
    - спецификации
    - сериализация объектов в ответ.

    Регистрацию обязательно проводить ПОСЛЕ регистрации ресурсов в приложении,
    иначе SpecTree не увидит URL-ы ресурсов.
    """
    register_error_handlers(app)
    register_specifications(app)
    register_serializer(app)
    return app



import inspect

import falcon.media
from falcon import Request, Response

from classic.components import component
from classic.error_handling import Error, ErrorsList

from defspec import OpenAPI, RenderTemplate
import msgspec

from . import error_handlers


def is_method_with_spec(fn):
    return callable(fn) and hasattr(fn, '__specification__')


class OpenAPIResource:
    def __init__(self, openapi) -> None:
        self.openapi = openapi
        self.spec = None

    def on_get(self, req: Request, resp: Response):
        if not self.spec:
            self.spec = self.openapi.to_json()
        resp.content_type = falcon.MEDIA_JSON
        resp.data = self.spec


class OpenAPIRender:
    def __init__(self, spec_url: str, template: RenderTemplate) -> None:
        self.template = template.value.format(spec_url=spec_url)

    def on_get(self, req: Request, resp: Response):
        resp.content_type = falcon.MEDIA_HTML
        resp.text = self.template


@component(init=False)
class App(falcon.App):
    
    def __init__(
        self,
        openapi: bool = False,
        open_api_spec_url: str = '/openapi/spec.json',
        swagger_url: str = '/openapi/swagger',
        media_type=falcon.DEFAULT_MEDIA_TYPE,
        request_type=Request,
        response_type=Response,
        middleware=None,
        router=None,
        independent_middleware=True,
        cors_enable=False,
        sink_before_static_route=True,
    ):
        super().__init__(
            media_type,
            request_type,
            response_type,
            middleware,
            router,
            independent_middleware,
            cors_enable,
            sink_before_static_route,
        )

        if openapi:
            self.openapi = OpenAPI()
            self.add_route(open_api_spec_url, OpenAPIResource(self.openapi))
            self.add_route(swagger_url, OpenAPIRender(
                open_api_spec_url,
                RenderTemplate.SWAGGER,
            ))
        else:
            self.openapi = None

        self.req_options.auto_parse_qs_csv = True
        self.req_options.keep_blank_qs_values = False

        self.resp_options.media_handlers[
            falcon.MEDIA_JSON
        ] = falcon.media.JSONHandler(
            dumps=msgspec.json.encode,
        )

        self.add_error_handler(Error, error_handlers.app_error)
        self.add_error_handler(ErrorsList, error_handlers.app_errors_list)
        self.add_error_handler(
            msgspec.ValidationError,
            error_handlers.validation_error,
        )

    def add_route(self, uri_template, resource, **kwargs):
        if self.openapi:
            for name, handler in inspect.getmembers(
                resource, is_method_with_spec,
            ):
                self.openapi.register_route(
                    uri_template, name[3:], **handler.__specification__
                )
        return super().add_route(uri_template, resource, **kwargs)



from functools import wraps
from typing import Type, Optional

from classic.components import doublewrap
from falcon import Request, Response
import msgspec


@doublewrap
def specification(
    method,
    summary: Optional[str] = None,
    media: Optional[Type] = None,
    response: Optional[Type] = None,
    query: Optional[Type] = None,
    header: Optional[Type] = None,
    cookie: Optional[Type] = None,
    deprecated: bool = False,
):
    @wraps(method)
    def wrapper(resource, req: Request, resp: Response, **kwargs):
        if query:
            req.context.query = msgspec.convert(
                req.params, type=query, strict=False,
            )
        if header:
            req.context.headers = msgspec.convert(
                req.headers, type=header, strict=False,
            )
        if cookie:
            req.context.cookies = msgspec.convert(
                req.cookies, type=cookie, strict=False,
            )
        if media:
            req.context.media = msgspec.json.decode(
                req.bounded_stream.read(),
                type=media,
            )
        method(resource, req, resp, **kwargs)

    wrapper.__specification__ = dict(
        summary=summary,
        request_type=media,
        response_type=response,
        query_type=query,
        header_type=header,
        cookie_type=cookie,
        deprecated=deprecated,
    )

    return wrapper



from falcon import Request, Response
from classic.components import component
from classic.http_api import App, specification
import msgspec


# Описывает параметры запроса для GET /api/some_obj
class SomeObjFilter(msgspec.Struct):
    number: int


# Описывает структуру ответа
class SomeObj(msgspec.Struct):
    some_attr: int


# Описывает структуру запроса для POST /api/some_obj
class CreateSomeObjRequest(msgspec.Struct):
    some_attr: int


@component
class SomeObjResource:

    @specification(query=SomeObjFilter, response=SomeObj)
    def on_get(self, request: Request, response: Response):
        # Представим себе, что объекты берутся из БД
        response.media = [
            SomeObj(number)
            for number in range(
                # Объект запроса содержится в контексте под именем query.
                request.context.query.number,
            )
        ]

    @specification(media=CreateSomeObjRequest, response=SomeObj)
    def on_post(self, request: Request, response: Response):
        # Представим себе, что объект был сохранен в БД;)
        response.media = SomeObj(
            **msgspec.structs.asdict(request.context.media)
        )


# Композит
if __name__ == '__main__':
    from wsgiref.simple_server import make_server

    app = App(openapi=True)
    app.add_route('/api/some_obj', SomeObjResource())

    # 
    with make_server('', 8000, app) as httpd:
        httpd.serve_forever()



from .specification import specification, register_specifications
from .error_handlers import register_error_handlers
from .serialization import register_serializer
from .registration import register_all



from falcon import Request

from .base import Authenticator, AuthenticationFailed


class FallbackAuthenticator(Authenticator):

    def __init__(self, authenticators: list[Authenticator], **kwargs):
        super().__init__(**kwargs)
        self.authenticators = authenticators

    def __call__(self, request: Request) -> object:
        identity = None
        for authenticator in self.authenticators:
            try:
                identity = authenticator(request)
            except AuthenticationFailed:
                continue

        if identity is None:
            raise AuthenticationFailed()

        if self.factory:
            return self.factory(identity)



from typing import Any, Sequence, Optional

from falcon import Request

try:
    import jwt
except ImportError:
    pass


from .base import Authenticator, AuthenticationFailed


class JWTAuthenticator(Authenticator):

    def __init__(
        self,
        secret_key: str,
        field: str = 'sub',
        algorithms: Sequence[str] = None,
        decoding_options: Optional[dict[str, Any]] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.secret_key = secret_key
        self.algorithms = algorithms or ['HS256']
        self.decoding_options = decoding_options
        self.field = field

        try:
            jwt
        except NameError as error:
            raise ImportError(f'Package {error.name} should be installed')

    def __call__(self, request: Request) -> None:
        token = self._extract_token(request)
        request.context.jwt = self._decode_token(token)
        self._set_identity(request)

    def _set_identity(self, request: Request):
        request.context.identity = self.factory(
            request.context.jwt[self.field]
        )

    def _extract_token(self, request: Request) -> str:
        try:
            return request.auth.split(' ')[-1]
        except KeyError:
            raise AuthenticationFailed(
                'Authorization header must contain 2 value, '
                'splitted by whitespace'
            )
        except AttributeError:
            raise AuthenticationFailed('No Authorization header')

    def _decode_token(self, token: str):
        try:
            return jwt.decode(
                token,
                self.secret_key,
                algorithms=self.algorithms,
                options=self.decoding_options,
            )
        except jwt.DecodeError:
            raise AuthenticationFailed('Token decoding error')
        except jwt.PyJWTError as e:
            raise AuthenticationFailed(f'Unexpected token error [{str(e)}]')



from functools import partial

import orjson
from typing import Any

import falcon.media
import pydantic


def default(self, obj: Any):
    if isinstance(obj, pydantic.BaseModel):
        return obj.model_dump_json()
    raise TypeError


def register_serializer(app: falcon.App):
    """
    Регистрация orjson в качестве дефолтного сериализатора.
    Сериализация также будет упаковывать модели pydantic.
    """
    orjson_handler = falcon.media.JSONHandler(
        dumps=partial(orjson.dumps, default=default),
        loads=orjson.loads,
    )
    app.req_options.media_handlers[falcon.MEDIA_JSON] = orjson_handler
    app.resp_options.media_handlers[falcon.MEDIA_JSON] = orjson_handler
    return app



from typing import Any

from falcon import Request, Response, status_codes
import pydantic

from classic.error_handling import Error, ErrorsList


def serialize_pydantic_error(
    request: Request, response: Response,
    error: pydantic.ValidationError, params: dict[str, Any],
):
    response.status = status_codes.HTTP_400
    response.media = str(error)


def serialize_app_error(
    request: Request, response: Response,
    error: Error, params: dict[str, Any],
):
    response.status = status_codes.HTTP_422
    response.media = [{'type': error.code,
                       'msg': error.message,
                       'ctx': error.context}]

def serialize_errors_list(
    request: Request, response: Response,
    error: ErrorsList, params: dict[str, Any],
):
    response.status = status_codes.HTTP_422
    response.media = [
        {'type': e.code,
         'msg': e.message,
         'ctx': e.context}
        for e in error.errors
    ]


def register_error_handlers(app):
    """
    Регистрация обработчиков ошибок в приложении.
    """
    app.add_error_handler(
        pydantic.ValidationError,
        serialize_pydantic_error,
    )
    app.add_error_handler(Error, serialize_app_error)
    app.add_error_handler(ErrorsList, serialize_errors_list)
    return app
